/* Prosa generic command addons to tclsh for ET-Linux */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include <sys/mount.h>
#include <sys/types.h>
#include <sys/wait.h>

#ifdef __i386__ /* direct port I/O is only done on the x86 platform */
#  include <asm/io.h>
#  ifdef __GLIBC__
#    include <sys/perm.h>
# /* define KDSETLED   0x4B32  -- not used any more */
#  else
# /* include <linux/kd.h> -- not used any more */
#    include <linux/fs.h>
#  endif
#else
# /* define KDSETLED   0x4B32  -- not used any more */
#endif

#include "ettclInt.h"
#include "ettclPort.h"

#ifdef TCL_USE_PROSA_EXT


/*
 * buffer and procedures used by uuencode and uudecode
 */

#define UUBUFLEN 80
static unsigned char uubuffer[UUBUFLEN];
void uuencode_this(unsigned char *buffer, int c)
{
    static int ptr;
    static int count = 0;
    unsigned long l;

    if (ptr==0) ptr=1;
    if (c) {
	l = buffer[0]<<16 | buffer[1]<<8 | buffer[2];
	/* this uses '`' instead of ' '  */
	uubuffer[ptr++] = ' ' + (((l>>18)-1) & 0x3f) + 1;
	uubuffer[ptr++] = ' ' + (((l>>12)-1) & 0x3f) + 1;
	uubuffer[ptr++] = ' ' + (((l>> 6)-1) & 0x3f) + 1;
	uubuffer[ptr++] = ' ' + (((l>> 0)-1) & 0x3f) + 1;
	count += c;
    }
    if (count==45 || (count && c<3)) { /* if 0, don't print anything */
	uubuffer[0] = ' '+count;
	uubuffer[ptr] = '\0';
	printf("%s\n",uubuffer);
	count=0; ptr=1;
    }
}

void uudecode_this(FILE *f)
{
    unsigned long l;
    int ptr, count;
    if (uubuffer[0]=='`') return;
    count = uubuffer[0]-' ';
    for (ptr = 1; count; ptr+=4) {
	l = (((uubuffer[ptr]-' ')&0x3f) << 18)
	    | (((uubuffer[ptr+1]-' ')&0x3f) << 12)
	    | (((uubuffer[ptr+2]-' ')&0x3f) << 6)
	    | ((uubuffer[ptr+3]-' ')&0x3f);
	if (count) {putc(l>>16, f); count--;}
	if (count) {putc((l>>8) & 0xff , f); count--;}
	if (count) {putc(l&0xff, f); count--;}
    }
}

/*
 * Simple commands of general utility
 */

int Tcl_ProsaCmdCmd(ClientData dummy, Tcl_Interp *interp, 
		    int argc, char **argv)
{
    int i, j;
    static char string[16], dir[16], fstype[16];
    extern char globalargv0[];
    int fd;
    FILE *f;

    switch(argv[0][0]) {

        /*
         * "inp" and "inw" receive a port number on cmdline (hex or dec)
         * and returns the current input value.
         */
      case 'i':                                         /* in[pw] */
        if (argc!=2) goto wna;
        if (Tcl_GetInt(interp, argv[1], &i) != TCL_OK) {
            Tcl_AppendResult(interp, "\nwhile parsing a port number",
                             NULL);
            return TCL_ERROR;
        }
#ifdef __i386__ /* use direct instruction, but require access first */
        if (ioperm(i,2,1)) { /* ask for two ports, in case inw is used */
            Tcl_AppendResult(interp, argv[0], ": ioperm(): ", strerror(errno),
                             NULL);
            return TCL_ERROR;
        }
	if (argv[0][2]=='p') { /* simple 8-bit port */
            j = inb(i);
            sprintf(string,"0x%02x",j);
        } else {
            j = inw(i);
            sprintf(string,"0x%04x",j);
        }
#else /* not 386, use /dev/port */
	fd = open("/dev/port", O_RDONLY);
	if (fd < 0) {
            Tcl_AppendResult(interp, argv[0], ": /dev/port: ", strerror(errno),
                             NULL);
            return TCL_ERROR;
        }
	lseek(fd, i, SEEK_SET);
	if (argv[0][2]=='p') { /* simple 8-bit port */
	    unsigned char c;
            read(fd, &c, 1); j = c;
            sprintf(string,"0x%02x",j);
        } else {
	    unsigned short w;
            read(fd, &w, 2); j = w;
            sprintf(string,"0x%04x",j);
        }
	close(fd);
#endif
        Tcl_SetResult(interp,string,TCL_VOLATILE);
        return TCL_OK;

        /*
         * "outp" and "outtw" receive a port number and a byte/word value
         * and calls an outb or outw operation.
         */
      case 'o':                                         /* outp */
        if (argc!=3) goto wna;
        if (Tcl_GetInt(interp, argv[1], &i) != TCL_OK) {
            Tcl_AppendResult(interp, "\nwhile parsing a port number",
                             NULL);
            return TCL_ERROR;
        }
        if (Tcl_GetInt(interp, argv[2], &j) != TCL_OK) {
            Tcl_AppendResult(interp, "\nwhile parsing a byte/word value",
                             NULL);
            return TCL_ERROR;
        }
#ifdef __i386__
        if (ioperm(i,2,1)) { /* ask for twwo bytes, in case outw is used */
            Tcl_AppendResult(interp, argv[0], ": ioperm(): ", strerror(errno),
                             NULL);
            return TCL_ERROR;
        }
        if (argv[0][3]=='b') 
            outb(j,i);
        else
            outw(j,i);
#else /* not __i386__ */
	fd = open("/dev/port", O_RDWR);
	if (fd < 0) {
            Tcl_AppendResult(interp, argv[0], ": /dev/port: ", strerror(errno),
                             NULL);
            return TCL_ERROR;
        }
	lseek(fd, i, SEEK_SET);
	if (argv[0][2]=='p') { /* simple 8-bit port */
	    unsigned char c;
	    c = j; write(fd, &c, 1);
        } else {
	    unsigned short w;
            w = j; write(fd, &w, 2);
        }
	close(fd);
#endif
        return TCL_OK;

        /*
         * "readb", "readw", "readl" receive a memory address on
         * cmdline (hex or dec) and return the current content.
	 */

      case 'r':                                         /* read[bwl] */
        if (argc!=2) goto wna;
        if (Tcl_GetInt(interp, argv[1], &i) != TCL_OK) {
            Tcl_AppendResult(interp, "\nwhile parsing a memory address",
                             NULL);
            return TCL_ERROR;
        }
	fd = open("/dev/mem", O_RDONLY);
	if (fd < 0) {
            Tcl_AppendResult(interp, argv[0], ": /dev/mem: ", strerror(errno),
                             NULL);
            return TCL_ERROR;
        }
	lseek(fd, i, SEEK_SET);
	if (argv[0][4]=='b') {
	    unsigned char c;
            read(fd, &c, 1); j = c;
            sprintf(string,"0x%02x",j);
        } else if (argv[0][4]=='w') {
	    unsigned short w;
            read(fd, &w, 2); j = w;
            sprintf(string,"0x%04x",j);
        } else /* long */ {
	    unsigned int l; /* long may be 64-bit. Int is 32, hopefully */
            read(fd, &l, 4); j = l;
            sprintf(string,"0x%08x",j);
	}
	close(fd);
        Tcl_SetResult(interp,string,TCL_VOLATILE);
        return TCL_OK;

        /*
         * "writeb", "writew" and "writel" receive an address and a value,
	 * and write that value to memory.
         */
      case 'w':                                         /* write[bwl] */
        if (argc!=3) goto wna;
        if (Tcl_GetInt(interp, argv[1], &i) != TCL_OK) {
            Tcl_AppendResult(interp, "\nwhile parsing a port number",
                             NULL);
            return TCL_ERROR;
        }
        if (Tcl_GetInt(interp, argv[2], &j) != TCL_OK) {
            Tcl_AppendResult(interp, "\nwhile parsing a byte/word value",
                             NULL);
            return TCL_ERROR;
        }
	fd = open("/dev/mem", O_RDWR);
	if (fd < 0) {
            Tcl_AppendResult(interp, argv[0], ": /dev/mem: ", strerror(errno),
                             NULL);
            return TCL_ERROR;
        }
	lseek(fd, i, SEEK_SET);
	if (argv[0][4]=='b') {
	    unsigned char c = j;
            write(fd, &c, 1);
        } else if (argv[0][4]=='w') {
	    unsigned short w = j;
            write(fd, &w, 2);
        } else /* long */ {
	    unsigned int l=j; /* long may be 64-bit. Int is 32, hopefully */
            write(fd, &l, 4);
	}
	close(fd);
        return TCL_OK;


#if 0 /* setleds was an horrible hack. Removed */

        /*
         * "setleds" is passed a file handle and a number, and calls ioctl()
         * to toggle the lights 
         */
      case 's':                                         /* setleds */
        if (argc!=3) goto wna;

        if (Tcl_GetInt(interp, argv[2], &i) != TCL_OK) {
            Tcl_AppendResult(interp, "\nwhile parsing a port number",
                             NULL);
            return TCL_ERROR;
        }
        chanPtr = Tcl_GetChannel(interp, argv[1], NULL);
        if (chanPtr == (Tcl_Channel) NULL) {
            return TCL_ERROR;
        }
        fsPtr = Tcl_GetChannelInstanceData(chanPtr);
        if (fsPtr->inFile) {
            fd = (int) Tcl_GetFileInfo(fsPtr->inFile, NULL);
        } else {
            fd = (int) Tcl_GetFileInfo(fsPtr->outFile, NULL);
        }
        if (ioctl(fd,KDSETLED,(long)i)) {
            Tcl_AppendResult(interp, "setleds: ioctl(): ", strerror(errno),
                             NULL);
            return TCL_ERROR;
        }
        return TCL_OK;
#endif /* if 0 */

        /*
         * "newcmdname" should modify the "Name:" field in /proc/self/status
         */
      case 'n':                                         /* newcmdname */
        if (argc!=2) goto wna;
	strncpy(globalargv0,argv[1],15); /* not works */
        /*
         * Try to do it through /var/run instead
         */
        sprintf(string,"/var/run/%i",getpid());
        f=fopen(string,"w");
        if (f) {
            fprintf(f,"%s\n",argv[1]);
            fclose(f);
        }
        return TCL_OK;

        /*
         * "mount": no arguments to remount root, "mount proc" to mount /proc
         * The first three words in /etc/fstab must be "<blkdev> / <type>"
         */
      case 'm':                                         /* mount */
        if (argc==1) {
            f=fopen("/etc/fstab","r");
            if (!f) {
                Tcl_AppendResult(interp, argv[0], ": /etc/fstab: ",
                                 strerror(errno), NULL);
                return TCL_ERROR;
            }
            i=fscanf(f,"%s %s %s",string,dir,fstype);
            fclose(f);
            if (i!=3) {
                Tcl_AppendResult(interp, argv[0], ": /etc/fstab: wrong data",
                                 NULL);
                return TCL_ERROR;
            }
            if (mount(string, dir, fstype,MS_REMOUNT | MS_MGC_VAL, NULL)) {
                Tcl_AppendResult(interp, argv[0], ": mount(): ",
                                 strerror(errno), NULL);
                return TCL_ERROR;
            }
            return TCL_OK;
        }
        if (argc==2) {
            if (mount("none","/proc","proc", MS_MGC_VAL, NULL)) {
                Tcl_AppendResult(interp, argv[0], ": mount(): ",
                                 strerror(errno), NULL);
                return TCL_ERROR;
            }
        return TCL_OK;
        }
        goto wna;

      case 'u':
	if (argv[0][1]=='m') {                /* umount */
	    sync();
	    umount("/proc");
	    umount("/");
	    return TCL_OK;
	}
	if (argv[0][2]=='e') {                /* uuencode */
	    unsigned char buffer[4];
	    if (argc!=2) goto wna;
	    if (!(f=fopen(argv[1],"r"))) {
		Tcl_AppendResult(interp, argv[0], ": ",argv[1],": ",
				 strerror(errno), NULL);
		return TCL_ERROR;
	    }
	    printf("begin 644 %s\n",argv[1]);
	    while ( (i=fread(buffer,1,3,f)) > 0)
		uuencode_this(buffer, i);
	    uuencode_this(buffer, 0);
	    printf("`\nend\n");
	    fflush(stdout); /* just in case... */
	    return TCL_OK;
	}
	if (argv[0][2]=='d') {                /* uudecode */
	    int mode;
	    char filename[UUBUFLEN];

	    f = NULL;
	    while (fgets(uubuffer, UUBUFLEN, stdin)) {
		if (!strncmp(uubuffer, "end", 3)) {
		    if (f) {
			fclose(f);
			chmod(filename,mode);
		    }
		    return TCL_OK;
		}
		if (!strncmp(uubuffer, "begin", 5)) {
		    if (sscanf(uubuffer, "begin %o %s",&mode,filename)!=2) {
			Tcl_AppendResult(interp, argv[0], ": error in data\n",
					 NULL);
			return TCL_ERROR;
		    }
		    if (!(f=fopen(filename,"w"))) {
			Tcl_AppendResult(interp, argv[0], ": ",uubuffer,": ",
					 strerror(errno), NULL);
			return TCL_ERROR;
		    }
		    continue;
		}
		if (!f) continue;
		uudecode_this(f);
	    }
	    if (f) {
		Tcl_AppendResult(interp, argv[0], ": incomplete input,\"",
				 filename, "\" is corrupted",NULL);
		fclose(f);
		return TCL_ERROR;
	    }
	    return TCL_OK;
	}


      default:
        Tcl_SetResult(interp, "Unknown command", TCL_STATIC);
        return TCL_ERROR;
    }

  wna:
    Tcl_AppendResult(interp, argv[0], ": Wrong number of arguments\n",
                     NULL);
    return TCL_ERROR;
}

int Tcl_ProsaTimeCmd(ClientData dummy, Tcl_Interp *interp, 
		    int argc, char **argv)
{
    /*
     * The command behaves in the following way:
     *
     *    xtime                           returns the time as %li
     *    xtime -milli                    returns the time as %li.%03li
     *    xtime -micro                    returns the time as %li.%06li
     *    xtime -format <fmt> [<time_t>]  returns the time as strftime(<fmt>)
     *    xtime -http [<time_t>]          uses the std http format and GMT
     *    xtime -diff <time>      returns the difference (same fmt as <time>)
     *    xtime -timeto <time>    returns the difference in milliseconds
     *    xtime -wait <tvar> <step>       waits untile $tvar+step, updates tvar
     *    xtime -split <array> [<time_t>] splits the date in components
     *    xtime -join <array>             joins components into a time_t
     */

    time_t  t=0;
    long tdiff, milli, micro, micro2;
    unsigned long sec, sec2;
    struct timeval tv, tvstep;
    struct  tm      *loc;
    unsigned char *fmt = NULL;
    char *var;
    int args, res;
    /* as a (dirty) shortcut, use an int[] for struct tm in split/join*/
    int *tm = (int *)loc;
    char *tmnames[]={"sec", "min", "hour", "mday", "mon", "year",
		   "wday", "yday", "isdst", NULL};
    #define TMNAMES 10

    enum mode {
	TIME_T = 0,
	TIME_MILLI, TIME_MICRO, TIME_FORMAT, TIME_FORMATNOW, TIME_HTTP,
	TIME_HTTPNOW, TIME_DIFF, TIME_TIMETO, TIME_WAIT,
	TIME_SPLIT, TIME_SPLITNOW, TIME_JOIN
    };

    struct {
	char *opt;
	int mode;
	int narg;
    } *optr, optionlist[] = {
	{"-milli",    TIME_MILLI,     2},
	{"-micro",    TIME_MICRO,     2},
	{"-format",   TIME_FORMATNOW, 3},
	{"-format",   TIME_FORMAT,    4},
	{"-http",     TIME_HTTP,      3},
	{"-http",     TIME_HTTPNOW,   2},
	{"-diff",     TIME_DIFF,      3},
	{"-timeto",   TIME_TIMETO,    3},
	{"-wait",     TIME_WAIT,      4},
	{"-split",    TIME_SPLITNOW,  3},
	{"-split",    TIME_SPLIT,     4},
	{"-join",     TIME_JOIN,      3},
	{NULL,        -1,             0},
    };

    int mode = TIME_T;

    if (argc>1) {
	for (optr = optionlist; optr->opt; optr++)
	    if (!strcmp(argv[1], optr->opt) && optr->narg == argc)
		break;
	if (!optr->opt) {
	    Tcl_SetResult(interp, "xtime: wrong arguments\n", TCL_STATIC);
	    return TCL_ERROR;
	}
	mode = optr->mode;
    }

    /* use uubuffer for the return value */
    switch(mode) {
        case TIME_FORMAT:
	    t = (time_t)atol(argv[3]);
        case TIME_FORMATNOW:
	    fmt = argv[2]; /* and fall through */
        case TIME_T:
	    if (!t) t = time(0);
	    if (fmt) {
		loc = localtime(&t);
		strftime(uubuffer, UUBUFLEN, fmt, loc);
	    } else {
	      sprintf(uubuffer,"%ld",(long)t);
	    }		
	    break;

       case TIME_HTTPNOW:
	   t = time(0); /* and fall through */
       case TIME_HTTP:
	   if (!t) t = (time_t)atol(argv[2]);
	   loc = gmtime(&t);
	   fmt = "%a, %d %b %Y %H:%M:%S GMT";
	   strftime(uubuffer, UUBUFLEN, fmt, loc);
	   break;
	   

       case TIME_MILLI:
       case TIME_MICRO:
	   gettimeofday(&tv, NULL);
	   if (mode == TIME_MILLI) {
	       sprintf(uubuffer, "%ld.%03ld",
		       (long)tv.tv_sec, (long)tv.tv_usec/1000);
	   } else {
	       sprintf(uubuffer, "%ld.%06ld", 
		       (long)tv.tv_sec, (long)tv.tv_usec);
	   }
	   break;

       case TIME_TIMETO: /* always returns in millisecs (can't be negative) */
       case TIME_DIFF:   /* returns same fmt as input (can be negative) */
	   milli = micro = 0;
	   args = sscanf(argv[2],"%ld.%3ld%3ld", &sec, &milli, &micro);
	   if (args<1)
	       goto xtimeerr;
	   /* now convert it in sec.micro */
	   micro = milli * 1000 + micro;
	   /* and now get the difference into tdiff.micro (sec is unsigned) */
	   gettimeofday(&tv, NULL);
	   tdiff = sec - tv.tv_sec;
	   micro = micro - tv.tv_usec;
	   if (micro < 0) {
	       micro += 1000 * 1000;
	       tdiff--;
	   }
	   /*
	    * now, micro is always 0-999999, and tdiff may be negative.
	    * A time of -1.5 is thus represented by tdiff==-2 and micro==5e5
	    * (tdiff is the floor of the decimal time-difference
	    */

	   if (mode == TIME_TIMETO) {
	       milli = tdiff * 1000 + micro/1000;
	       /* a negative time is 0, so no problems with signs */
	       if (milli < 0) milli = 0;
	       sprintf(uubuffer, "%ld", milli);
	       break;
	   }

	   /* This its "-diff", and we have problems with negatives */
	   switch(args) {
	       case 1:  /* print only the seconds as floor(timeandmicro) */
		   sprintf(uubuffer, "%ld", tdiff);
		   break;
		   
	       case 2: /* print seconds and milli */
		   if (tdiff == -1 && micro >= 1000) {
		       /* special case: "-0" is not a conventional int */
		       sprintf(uubuffer, "-0.%03ld", 1000 - micro/1000);
		       break;
		   }
		   if (tdiff < 0 && micro >= 1000) {
		       tdiff++;
		       micro = 1000000 - micro;
		   }
		   sprintf(uubuffer, "%ld.%03ld", tdiff, micro/1000);
		   break;

	       case 3: /* print seconds and micro */
		   if (tdiff == -1 && micro) {
		       /* special case: "-0" is not a conventional int */
		       sprintf(uubuffer, "-0.%06ld", 1000000 - micro);
		       break;
		   }
		   if (tdiff < 0 && micro) {
		       tdiff++;
		       micro = 1000000 - micro;
		   }
		   sprintf(uubuffer, "%ld.%06ld", tdiff,micro);
		   break;
	   }
	   break;

       case TIME_SPLITNOW:
	   t = time(0); /* and fall through */
       case TIME_SPLIT:
	   if (!t) t = (time_t)atol(argv[3]);
	   loc = localtime(&t); tm = (int *)loc; res = 0;
	   loc->tm_year += 1900;
	   /* use uubuffer as temporary storage, as usual. "res" is errorfl */
	   for (mode=0; tmnames[mode]; mode++) {
	       sprintf(uubuffer, "%i", tm[mode]);
	       res += !Tcl_SetVar2(interp, argv[2], tmnames[mode], uubuffer, 0);
	   }
	   if (res) {
	       Tcl_SetResult(interp, "time -split: error in setting vars",
			     TCL_STATIC);
	       return TCL_ERROR;
	   }
	   uubuffer[0]='\0'; /* no result */
	   break;

       case TIME_JOIN:
	   tm = calloc(TMNAMES, sizeof(int));
	   sec = (unsigned long)-1;
	   if (tm) {
	       for (mode = 0; tmnames[mode]; mode++) {
		   fmt = Tcl_GetVar2(interp, argv[2], tmnames[mode], 0);
		   if (fmt)
		       sscanf(fmt,"%i",&tm[mode]);
	       }
	       loc = (struct tm *)tm;
	       if (loc->tm_year >= 1900) loc->tm_year -= 1900;
	       sec = mktime(loc);
	       free(tm);
	   }
	   if (sec == (unsigned long)-1) {
	       Tcl_SetResult(interp,"xtime -join: nomem or other error",
			     TCL_STATIC);
	       return TCL_ERROR;
	   }
	   sprintf(uubuffer,"%li",sec);
	   break;
	   
       case TIME_WAIT:
	   /* 
	    * The "-wait" functionality is meant to help in implementing
	    * repeated actions. It gets two arguments: the name of
	    * a variable and a time difference. The command will wait
	    * till time "$var + diff" and then return. It also updates var
	    * so that the function can be used the next time. Both the
	    * variable and the time difference can be either seconds
	    * (as returned by "xtime"), or milliseconds (as returned by
	    * "xtime -milli") or microseconds (as returned by "xtime -micro").
	    * However, the number of decimal digits in both items must be
	    * the same (if they are not, the variable's one takes precedence
	    *, and the result in undefined (read: "wrong") if the decimals are
	    * anything but 0, 3, 6 digits.
	    *
	    * If the time $var+diff is in the past, the command will
	    * return immediately. If even the next time event ($var + 2*diff)
	    * is in the past, the var will be incremented twice and so on.
	    * On exit from the command, thus, the variable will represent
	    * the most recent timestamp in the past. The return value is the
	    * number of "missed" events: 0 in normal case, and 1 or more
	    * if the command is invoked so late that some even has been
	    * missed.  Thus, the application developer can choose to
	    * either loose missed events or process them late.
	    *
	    * Loosing events:
	    *                  while 1 {
	    *                      xtime -wait var $interval
	    *                      <do stuff>
	    *                  }
	    * 
	    * Late processing:
	    *                  while 1 {
	    *                      set i [xtime -wait var $interval]
	    *                      for {} {$i>=0} {incr i -1} {
	    *                          <do stuff>
	    *                      }
	    *                  }
	    */

	   if (argc != 4) goto xtimeerr;
	   var = Tcl_GetVar(interp, argv[2], TCL_LEAVE_ERR_MSG);
	   if (!var) return TCL_ERROR; /* error already there */

	   /* build tvstep */
	   milli = micro = 0;
	   args = sscanf(argv[3],"%ld.%3ld%3ld", &sec, &milli, &micro);
	   if (args < 1) goto xtimeerr;
	   tvstep.tv_sec = sec;
	   tvstep.tv_usec = milli * 1000 + micro;
	   
	   /* retrieve original time */
	   milli = micro = 0;
	   args = sscanf(var,"%ld.%3ld%3ld", &sec, &milli, &micro);
	   if (args < 1) goto xtimeerr;
	   micro = milli * 1000 + micro;
	   /* increment once */
	   sec += tvstep.tv_sec;
	   micro += tvstep.tv_usec;
	   if (micro >= 1000 * 1000) {
	       micro -= 1000 * 1000;
	       sec++;
	   }

	   res = 1;
	   while (1) {
	       /* retrieve current time */
	       gettimeofday(&tv, NULL);
	       if (tv.tv_sec > sec) break;
	       if (tv.tv_sec == sec && tv.tv_usec >= micro) break;

	       /* sleep and return */
	       tv.tv_sec = sec - tv.tv_sec;
	       tv.tv_usec = micro - tv.tv_usec;
	       if (tv.tv_usec < 0) {
		   tv.tv_usec += 1000 * 1000;
		   tv.tv_sec --;
	       }
	       if (select(0, NULL, NULL, NULL, &tv)<0) {
		   /* accept EINTR, as we might just have reaped a child */
		   if (errno != EINTR) {
		       Tcl_ResetResult(interp);
		       Tcl_AppendResult(interp, "xtime -wait: ",
					strerror(errno), NULL);
		       return TCL_ERROR;
		   }
		   Tcl_AsyncInvoke(NULL, 0);
		   continue;
	       }
	       goto xtimewait_ok;
	    }

	   /*
	    * no need to sleep, but count missed events
	    * here, tv is the current time, tvstep the step,
	    * sec:micro is the time of the first (missed) event
	    * and is also the result to be set in the variable, that is,
	    * the last event we missed.
	    */
	   {
	       /* use a long long to avoid overruns */
	       long long diff; long rest;
	       diff = 1000*1000 * (tv.tv_sec - sec) + tv.tv_usec - micro;
	       res = 1 + diff/(tvstep.tv_sec * 1000*1000 + tvstep.tv_usec);
	       rest = diff%(tvstep.tv_sec * 1000*1000 + tvstep.tv_usec);
	       sec = tv.tv_sec - rest/(1000*1000);
	       micro = tv.tv_usec - rest%(1000*1000);
	       if (micro < 0) {
		   micro += 1000*1000;
		   sec -= 1;
	       }
	   }

        xtimewait_ok:
	   switch(args) {
	       case 1:  /* print only the seconds */
		   sprintf(uubuffer, "%ld", sec);
		   break;
		   
	       case 2: /* print seconds and milli */
		   sprintf(uubuffer, "%ld.%03ld", sec, micro/1000);
		   break;

	       case 3: /* print seconds and micro */
		   sprintf(uubuffer, "%ld.%06ld", sec,micro);
		   break;
	   }
	   Tcl_SetVar(interp, argv[2], uubuffer, 0);
	   sprintf(uubuffer, "%i", res);
    }
    Tcl_SetResult(interp,uubuffer,TCL_VOLATILE);
    return TCL_OK;

 xtimeerr:
    /* argv[1] is valid, as the argc==0 case never jums here */
    Tcl_AppendResult(interp, "xtime ", argv[1], ": Wrong arguments\n", NULL);
    return TCL_ERROR;
}


#endif
